# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import cast

from mkdocs.config import config_options as option
from mkdocs.config.base import Config

from mkdocs_publisher._shared.mkdocs_utils import ConfigChoiceEnum


class SlugModeChoiceEnum(ConfigChoiceEnum):
    TITLE = 0, True, False
    FILENAME = 1, False, False


class PublishChoiceEnum(ConfigChoiceEnum):
    DRAFT = 0, False, False
    HIDDEN = 1, False, False
    PUBLISHED = 2, False, False
    TRUE = 3, False, True
    FALSE = 4, False, True

    @classmethod
    def drafts(cls) -> list:
        return cls._get_enums([cls.DRAFT, cls.FALSE])  # pragma: no cover

    @classmethod
    def published(cls) -> list:
        return cls._get_enums([cls.PUBLISHED, cls.TRUE])  # pragma: no cover


class OverviewChoiceEnum(ConfigChoiceEnum):
    AUTO = 0, True, False
    TRUE = 1, False, True
    FALSE = 2, False, True


class TitleChoiceEnum(ConfigChoiceEnum):
    META = 0, True, False
    HEAD = 1, False, False
    FILE = 2, False, False


class _MetaSlugConfig(Config):
    enabled = option.Type(bool, default=True)
    mode = option.Choice(
        choices=SlugModeChoiceEnum.choices(), default=SlugModeChoiceEnum.default()
    )
    warn_on_missing = option.Type(bool, default=True)
    key_name = option.Type(str, default="slug")


class _MetaOverviewConfig(Config):
    enabled = option.Type(bool, default=True)
    default = option.Choice(
        choices=OverviewChoiceEnum.choices(), default=OverviewChoiceEnum.default()
    )
    key_name = option.Type(str, default="overview")


class _MetaPublishConfig(Config):
    search_in_hidden = option.Type(bool, default=False)
    search_in_draft = option.Type(bool, default=False)
    file_default = option.Choice(choices=PublishChoiceEnum.choices(), default=False)
    file_warn_on_missing = option.Type(bool, default=True)
    dir_default = option.Choice(choices=PublishChoiceEnum.choices(), default=True)
    dir_warn_on_missing = option.Type(bool, default=False)
    key_name = option.Type(str, default="publish")


class _MetaTitleConfig(Config):
    key_name = option.Type(str, default="title")
    mode = option.Choice(choices=TitleChoiceEnum.choices(), default=TitleChoiceEnum.default())
    warn_on_missing_meta = option.Type(bool, default=True)
    warn_on_missing_header = option.Type(bool, default=True)


class MetaPluginConfig(Config):
    dir_meta_file = option.Choice(["README.md", "index.md"], default="README.md")

    overview: _MetaOverviewConfig = cast(
        _MetaOverviewConfig, option.SubConfig(_MetaOverviewConfig)
    )
    publish: _MetaPublishConfig = cast(_MetaPublishConfig, option.SubConfig(_MetaPublishConfig))
    slug: _MetaSlugConfig = cast(_MetaSlugConfig, option.SubConfig(_MetaSlugConfig))
    title: _MetaTitleConfig = cast(_MetaTitleConfig, option.SubConfig(_MetaTitleConfig))



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import re
from dataclasses import dataclass
from typing import cast

from mkdocs.config.defaults import MkDocsConfig
from mkdocs.structure.pages import Page

# noinspection PyProtectedMember
from mkdocs_publisher._shared import links
from mkdocs_publisher.blog.plugin import BlogPlugin

log = logging.getLogger("mkdocs.plugins.publisher.obsidian.backlinks")


@dataclass
class Link:
    text: str
    destination: str
    title: str
    source: str


class BacklinkLinks:
    def __init__(
        self,
        mkdocs_config: MkDocsConfig,
        backlinks: dict[str, list[Link]],
    ):
        self._mkdocs_config: MkDocsConfig = mkdocs_config
        self._backlinks: dict[str, list[Link]] = backlinks

    @staticmethod
    def _other_link_to_text(match: re.Match) -> str:
        """Return only a text from a link (http or markdown)"""
        return match.groupdict()["link"]

    @staticmethod
    def _create_backlink(match: re.Match):
        link = links.LinkMatch(**match.groupdict())
        return link.as_backlink

    def _parse_markdown_link(self, match: re.Match, page: Page, line: str):
        """Parse markdown link"""
        original_link = links.LinkMatch(**match.groupdict())
        backlink_link = f"{self._mkdocs_config.site_url}{page.url}{original_link.backlink_anchor}"

        # Convert document link into backlink
        link_replacement = f'<a href="{backlink_link}">{original_link.text}</a>'
        backlink_text = line.replace(str(original_link), link_replacement)

        # Convert other links into text
        backlink_text = re.sub(links.MD_LINK_RE, self._other_link_to_text, backlink_text)
        backlink_text = re.sub(links.HTTP_LINK_RE, self._other_link_to_text, backlink_text)

        # Create a backlink with context
        backlink_text = f'<p class="obsidian_backlink">{backlink_text}</p>'

        # Convert relative backlinks into absolute backlinks inside document directory
        original_link_source = page.file.src_uri
        if len(original_link.link.split("/")) == 1:
            destination_pre = original_link_source.split("/")[:-1]
            destination_pre.append(str(original_link.link))
            original_link.link = "/".join(destination_pre)
        original_link.link = original_link.link.replace("../", "")

        new_link = Link(
            text=backlink_text,
            destination=backlink_link,
            source=original_link.link,
            title=str(page.title),
        )

        # Get blog temporary files
        temp_blog_files = []
        if "pub-blog" in self._mkdocs_config.plugins:
            blog_plugin: BlogPlugin = cast(BlogPlugin, self._mkdocs_config.plugins["pub-blog"])
            temp_blog_files = blog_plugin.blog_config.temp_files_list

        # Do not add backlink if backlinks points to the same document
        if (
            original_link_source != original_link.link
            and original_link_source not in temp_blog_files
        ):
            log.debug(
                f"Found backlink to: {original_link.link}"
                f"{original_link.anchor if original_link.anchor is not None else ''}"
            )
            if original_link.link not in self._backlinks:
                self._backlinks[original_link.link] = [new_link]
            else:
                link_found = False
                for existing_link in self._backlinks[original_link.link]:
                    # Add backlink to an existing one
                    if existing_link.title == new_link.title:
                        link_found = True
                        existing_link.text = f"{existing_link.text}<hr>{new_link.text}"
                if not link_found:
                    self._backlinks[original_link.link].append(new_link)

    def find_markdown_links(self, markdown: str, page: Page):
        # """Find all markdown backlinks"""
        for line in markdown.split("\n"):
            for match in re.finditer(links.MD_LINK_RE, line):
                self._parse_markdown_link(match=match, page=page, line=line)

    def convert_to_backlink(self, markdown: str) -> str:
        """Convert backlink to link with an anchor for direct navigation after clicking on it"""
        return re.sub(links.MD_LINK_RE, self._create_backlink, markdown)



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import importlib.resources
import json
import logging
import os
import shutil
from pathlib import Path

import click

import mkdocs_publisher_docs
from mkdocs_publisher._shared import file_utils
from mkdocs_publisher._shared import mkdocs_utils

log = logging.getLogger("mkdocs.plugins.publisher.cli.dev")


@click.group
def app():
    """Development tools."""
    pass


def iter_res(src_dir: Path, dst_dir: Path, res_path: Path):
    for path in res_path.iterdir():
        if path.is_dir():
            iter_res(src_dir=src_dir, dst_dir=dst_dir, res_path=path)
        else:
            src_file = path
            dst_file = dst_dir / src_file.relative_to(src_dir)
            os.makedirs(name=str(dst_file.parents[0]), exist_ok=True)
            shutil.copy(src=src_file, dst=dst_file)


@app.command()
def docs_copy():
    mkdocs_config = mkdocs_utils.get_mkdocs_config()
    docs_path = Path().cwd() / mkdocs_config.docs_dir
    res_path = Path(str(importlib.resources.files(mkdocs_publisher_docs)))

    if docs_path == res_path:
        log.debug("Can't copy to the same directory!")
    else:
        if click.confirm(f"Remove content of '{docs_path}?'"):
            shutil.rmtree(docs_path, ignore_errors=True)
        iter_res(src_dir=res_path, dst_dir=docs_path, res_path=res_path)


@app.command()
def css_min():
    """Minify project CSS files."""
    project_dir = Path.cwd() / "mkdocs_publisher"
    for input_css_file in project_dir.rglob("*.css"):
        if ".min" not in input_css_file.suffixes:
            output_css_file = input_css_file.parent / f"{input_css_file.stem}.min.css"
            print(output_css_file)
            cmd = [
                "postcss",
                str(input_css_file),
                "-m",
                "--verbose",
                "-u",
                "cssnano",
                "postcss-svgo",
                "-o",
                str(output_css_file),
            ]
            file_utils.run_subprocess(cmd, capture_output=False)


@app.command()
def update_cov():
    """Update cov.json file based on full coverage.json file.

    File cov.json is used by shields.io to display % code coverage in a badge."""
    coverage_file = Path.cwd() / "coverage.json"
    cov_file = Path.cwd() / "cov.json"

    with coverage_file.open(mode="r") as coverage:
        coverage_data = json.load(coverage)["totals"]

    with cov_file.open(mode="w") as cov:
        json.dump(coverage_data, cov, indent=4)
        cov.write("\n")



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from mkdocs.config import config_options as option
from mkdocs.config.base import Config


class _MinifierCommonConfig(Config):
    cache_enabled = option.Type(bool, default=True)
    enabled = option.Type(bool, default=True)
    enabled_on_serve = option.Type(bool, default=False)
    exclude = option.Type(list, default=[])
    extensions = option.Type(list, default=[])


class _MinifierCssConfig(_MinifierCommonConfig):
    extensions = option.Type(list, default=[".[cC][sS][sS]"])
    postcss_path = option.Type(str, default="postcss")
    skip_minified = option.Type(bool, default=True)


class _MinifierJsConfig(_MinifierCommonConfig):
    extensions = option.Type(list, default=[".[jJ][sS]"])
    uglifyjs_path = option.Type(str, default="uglifyjs")
    skip_minified = option.Type(bool, default=True)


class _MinifierHtmlConfig(_MinifierCommonConfig):
    extensions = option.Type(list, default=[".[hH][tT][mM]", ".[hH][tT][mM][lL]"])
    html_minifier_path = option.Type(str, default="html-minifier")
    case_sensitive = option.Type(bool, default=True)
    minify_css = option.Type(bool, default=True)
    minify_js = option.Type(bool, default=True)
    remove_comments = option.Type(bool, default=True)
    remove_tag_whitespace = option.Type(bool, default=False)
    collapse_whitespace = option.Type(bool, default=True)
    conservative_collapse = option.Type(bool, default=True)
    collapse_boolean_attributes = option.Type(bool, default=True)
    preserve_line_breaks = option.Type(bool, default=True)
    sort_attributes = option.Type(bool, default=True)
    sort_class_name = option.Type(bool, default=True)
    max_line_length = option.Choice(
        [str(i) for i in range(80, 4097)], default="1024"
    )  # 0 - disabled


class _MinifierSvgConfig(_MinifierCommonConfig):
    extensions = option.Type(list, default=[".[sS][vV][gG]"])
    svgo_path = option.Type(str, default="svgo")
    multipass = option.Type(bool, default=True)


class _MinifierJpegConfig(_MinifierCommonConfig):
    extensions = option.Type(list, default=[".[jJ][pP][gG]", ".[jJ][pP][eE][gG]"])
    djpeg_path = option.Type(str, default="djpeg")
    cjpeg_path = option.Type(str, default="cjpeg")
    jpegtran_path = option.Type(str, default="jpegtran")
    optimise = option.Type(bool, default=True)
    progressive = option.Type(bool, default=True)
    copy_meta = option.Choice(["none", "comments", "icc", "all"], default="none")
    smooth = option.Choice([str(i) for i in range(0, 101)], default="10")  # 0 - disabled
    quality = option.Choice([str(i) for i in range(0, 101)], default="85")  # 0 - disabled


class _MinifierPngConfig(_MinifierCommonConfig):
    extensions = option.Type(list, default=[".[pP][nN][gG]"])
    pngquant_enabled = option.Type(bool, default=True)
    pngquant_path = option.Type(str, default="pngquant")
    pngquant_speed = option.Choice([str(i) for i in range(1, 12)], default="1")
    pngquant_quality = option.Choice([str(i) for i in range(0, 101)], default="95")  # 0 - disabled
    oxipng_enabled = option.Type(bool, default=True)
    oxipng_path = option.Type(str, default="oxipng")
    oxipng_max_compression = option.Type(bool, default=True)
    strip = option.Type(bool, default=True)


class MinifierConfig(Config):
    cache_enabled = option.Type(bool, default=True)
    cache_dir = option.Type(str, default=".pub_min_cache")
    cache_file = option.Type(str, default=".cached_files_list.yml")
    exclude = option.Type(list, default=[])
    threads = option.Type(int, default=0)  # 0 - default (read from system)

    js: _MinifierJsConfig = option.SubConfig(_MinifierJsConfig)  # type: ignore
    css: _MinifierCssConfig = option.SubConfig(_MinifierCssConfig)  # type: ignore
    jpeg: _MinifierJpegConfig = option.SubConfig(_MinifierJpegConfig)  # type: ignore
    png: _MinifierPngConfig = option.SubConfig(_MinifierPngConfig)  # type: ignore
    svg: _MinifierSvgConfig = option.SubConfig(_MinifierSvgConfig)  # type: ignore
    html: _MinifierHtmlConfig = option.SubConfig(_MinifierHtmlConfig)  # type: ignore



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import re
from collections import UserDict
from collections.abc import Generator
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Any
from typing import Optional
from urllib.parse import quote

from mkdocs.config.defaults import MkDocsConfig
from mkdocs.structure.files import File
from mkdocs.structure.files import Files
from mkdocs.utils import meta as meta_parser

# noinspection PyProtectedMember
from mkdocs_publisher._shared.urls import create_slug
from mkdocs_publisher.meta.config import MetaPluginConfig
from mkdocs_publisher.meta.config import OverviewChoiceEnum
from mkdocs_publisher.meta.config import PublishChoiceEnum
from mkdocs_publisher.meta.config import TitleChoiceEnum

log = logging.getLogger("mkdocs.plugins.publisher._shared.meta_files")


HEADINGS_RE = re.compile(r"^#+ (?P<title>[^|#\r\n\t\f\v]+)$")


@dataclass
class MetaFile:
    path: Path
    abs_path: Path
    is_dir: bool
    is_overview: bool = field(default=False)
    is_hidden: bool = field(default=False)
    is_draft: Optional[bool] = field(default=None)
    title: Optional[str] = field(default=None)
    slug: Optional[str] = field(default=None)
    file: Optional[File] = field(default=None)

    @property
    def name(self) -> str:
        return self.path.name

    @property
    def parent(self) -> Optional[Path]:
        parent = self.path.parent
        if str(parent) == ".":
            parent = None
        return parent


class MetaFiles(UserDict):
    def __init__(self):
        self._on_serve: bool = False
        self._mkdocs_config: Optional[MkDocsConfig] = None
        self._meta_plugin_config: Optional[MetaPluginConfig] = None
        self._hidden_paths: list[Path] = []
        super().__init__()

    @property
    def meta_file(self) -> str:
        return self._meta_plugin_config.dir_meta_file

    @property
    def on_serve(self) -> bool:
        return self._on_serve

    @on_serve.setter
    def on_serve(self, on_serve: bool):
        self._on_serve = on_serve

    def add_hidden_path(self, hidden_path: Optional[Path]):
        if hidden_path is not None:
            self._hidden_paths.append(hidden_path.relative_to(self._mkdocs_config.docs_dir))

    def set_configs(self, mkdocs_config: MkDocsConfig, meta_plugin_config: MetaPluginConfig):
        self._mkdocs_config = mkdocs_config
        self._meta_plugin_config = meta_plugin_config

    @staticmethod
    def _read_md_file(meta_file_path: Path) -> tuple[str, dict[str, Any]]:  # pragma: no cover
        with meta_file_path.open(encoding="utf-8-sig", errors="strict") as md_file:
            # Add empty line at the end of file and read all data
            return meta_parser.get_data(f"{md_file.read()}\n")

    def _get_title(self, meta_file: MetaFile, meta: dict[str, Any], markdown: str):
        title: Optional[str] = None
        mode = self._meta_plugin_config.title.mode
        if mode == TitleChoiceEnum.META:
            title = meta.get(self._meta_plugin_config.title.key_name)
            if (
                title is None
                and self._meta_plugin_config.title.warn_on_missing_meta
                and not (meta_file.is_draft or (meta_file.is_dir and not meta_file.is_overview))
            ):
                log.warning(
                    f'Title value from "{self._meta_plugin_config.title.key_name}" meta data '
                    f'is missing for file: "{str(meta_file.path)}"'
                )

        if title is None and (mode == TitleChoiceEnum.META or mode == TitleChoiceEnum.HEAD):
            headings = re.findall(HEADINGS_RE, markdown)
            title = str(headings[0]).strip() if len(headings) > 0 else None
            if (
                title is None
                and self._meta_plugin_config.title.warn_on_missing_header
                and not (meta_file.is_draft or (meta_file.is_dir and not meta_file.is_overview))
            ):
                log.warning(
                    f'Title value from first heading is missing for file: "{str(meta_file.path)}"'
                )

        if title is None and (mode == TitleChoiceEnum.META or mode == TitleChoiceEnum.FILE):
            title = str(meta_file.path.stem).replace("_", " ").title()

        meta_file.title = title

    def _get_slug(self, meta_file: MetaFile, meta: dict[str, Any]):
        meta_file.slug = create_slug(  # pragma: no cover
            file_name=meta_file.path.stem,
            slug_mode=self._meta_plugin_config.slug.mode,
            slug=meta.get(self._meta_plugin_config.slug.key_name),
            title=meta_file.title,
            warn_on_missing=self._meta_plugin_config.slug.warn_on_missing,
        )

    def _get_overview(self, meta_file: MetaFile, meta: dict[str, Any], markdown: str):
        """Overview works only for metafiles ("README.md", "index.md") that are detected as dir"""
        if not meta_file.is_dir:
            meta_file.is_overview = False
        else:
            is_overview = meta.get(
                self._meta_plugin_config.overview.key_name,
                self._meta_plugin_config.overview.default,
            )
            if is_overview == OverviewChoiceEnum.AUTO:
                meta_file.is_overview = bool(len(markdown.strip()))
            else:
                meta_file.is_overview = is_overview

    def _get_publish_status(self, meta_file: MetaFile, meta: dict[str, Any]):
        publish = meta.get(str(self._meta_plugin_config.publish.key_name), None)
        # Get default values if publish status is not specified
        if publish is None:
            if meta_file.is_dir:
                publish = self._meta_plugin_config.publish.dir_default
                if self._meta_plugin_config.publish.dir_warn_on_missing:
                    log.warning(
                        f'Missing "{self._meta_plugin_config.publish.key_name}" value in '
                        f'file "{meta_file.path}". Setting to '
                        f'default value: "{self._meta_plugin_config.publish.dir_default}".'
                    )
            else:
                publish = self._meta_plugin_config.publish.file_default
                if self._meta_plugin_config.publish.file_warn_on_missing:
                    log.warning(
                        f'Missing "{self._meta_plugin_config.publish.key_name}" value in '
                        f'file "{meta_file.path}". Setting to '
                        f'default value: "{self._meta_plugin_config.publish.file_default}".'
                    )

        # Override some values inherited from parent
        if meta_file.parent is not None:
            meta_file_parent: MetaFile = self[str(meta_file.parent)]
            if meta_file_parent.is_draft:
                publish = False
            if meta_file_parent.is_hidden:
                publish = PublishChoiceEnum.HIDDEN.name.lower()

        # When live preview is running, all pages are visible
        if self._on_serve and not meta_file.is_hidden:
            publish = True

        if publish not in PublishChoiceEnum.choices():
            publish = self._meta_plugin_config.publish.file_default
            log.warning(
                f'Wrong key "{self._meta_plugin_config.publish.key_name}" value '
                f'({publish}) in file "{meta_file.path}" (only '
                f"{PublishChoiceEnum.choices()} are possible)"
            )

        # Set values depends on publish status
        if meta_file.path in self._hidden_paths or publish == PublishChoiceEnum.HIDDEN:
            meta_file.is_hidden = True
            meta_file.is_draft = False
        elif publish in PublishChoiceEnum.drafts():
            meta_file.is_hidden = False
            meta_file.is_draft = True
        elif publish in PublishChoiceEnum.published():
            meta_file.is_hidden = False
            meta_file.is_draft = False

    def _get_metadata(self, meta_file: MetaFile, meta_file_path: Path):  # pragma: no cover
        """Read all metadata values for given file"""
        # Order of method execution is crucial for reading all values.
        markdown, meta = self._read_md_file(meta_file_path=meta_file_path)

        if self._meta_plugin_config.overview.enabled:
            self._get_overview(meta_file=meta_file, meta=meta, markdown=markdown)
        self._get_publish_status(meta_file=meta_file, meta=meta)
        self._get_title(meta_file=meta_file, meta=meta, markdown=markdown)
        self._get_slug(meta_file=meta_file, meta=meta)

    def __setitem__(self, path: str, meta_file: MetaFile):
        if meta_file.is_dir:
            meta_file_exists = False

            # Calculate properties based on meta file metadata
            meta_file_path = meta_file.abs_path.joinpath(self._meta_plugin_config.dir_meta_file)
            if meta_file_path.exists():
                meta_file_exists = True
                self._get_metadata(meta_file=meta_file, meta_file_path=meta_file_path)

            if not meta_file_exists:
                meta_file.title = str(meta_file.path.stem)
                meta_file.slug = create_slug(
                    file_name=str(meta_file.name),
                    slug_mode=self._meta_plugin_config.slug.mode,
                    slug=meta_file.path.stem,
                    title=meta_file.title,
                    warn_on_missing=self._meta_plugin_config.slug.warn_on_missing,
                )
                meta_file.is_draft = not self._meta_plugin_config.publish.dir_default
        else:
            self._get_metadata(meta_file=meta_file, meta_file_path=meta_file.abs_path)
        super().__setitem__(path, meta_file)

    def drafts(self, files: bool = False, dirs: bool = False) -> dict[str, MetaFile]:
        draft_files = {}
        for path, meta_file in self.items():
            if meta_file.is_draft and (
                (files and not dirs and not meta_file.is_dir)
                or (not files and dirs and meta_file.is_dir)
                or (not files and not dirs)
            ):
                draft_files[path] = meta_file
        return draft_files

    def hidden(self, files: bool = False, dirs: bool = False) -> dict[str, MetaFile]:
        hidden_files = {}
        for path, meta_file in self.items():
            if meta_file.is_hidden and (
                (files and not dirs and not meta_file.is_dir)
                or (not files and dirs and meta_file.is_dir)
                or (not files and not dirs)
            ):
                hidden_files[path] = meta_file
        return hidden_files

    def add_meta_files(self, ignored_dirs: list[Path]):
        """Iterate over all files and directories in docs directory"""
        for docs_file in sorted(Path(self._mkdocs_config.docs_dir).rglob("*")):
            meta_link: Optional[MetaFile] = None
            is_ignored = any(
                [docs_file.is_relative_to(ignored_dir) for ignored_dir in ignored_dirs]
            )

            if not is_ignored and docs_file.is_dir():
                meta_link = MetaFile(
                    path=docs_file.relative_to(self._mkdocs_config.docs_dir),
                    abs_path=docs_file,
                    is_dir=True,
                )
            elif (
                not is_ignored
                and docs_file.suffix == ".md"
                and docs_file.name != self._meta_plugin_config.dir_meta_file
            ):
                meta_link = MetaFile(
                    path=docs_file.relative_to(self._mkdocs_config.docs_dir),
                    abs_path=docs_file,
                    is_dir=False,
                )

            if meta_link is not None:
                self[str(meta_link.path)] = meta_link

    def change_files_slug(self, files: Files, ignored_dirs: list[Path]) -> Files:
        draft_files = self.drafts(files=True).keys()

        ignored_dirs.extend([d.abs_path for d in self.drafts(dirs=True).values()])

        new_files = Files(files=[])
        for file in files:
            file: File
            file_path: Path = Path(file.src_path)
            if (
                (
                    not any([Path(file.abs_src_path).is_relative_to(d) for d in ignored_dirs])
                    and file.src_path not in draft_files
                    and str(file_path.name) != self._meta_plugin_config.dir_meta_file
                )
                or (
                    str(file_path.name) == self._meta_plugin_config.dir_meta_file
                    and str(file_path.parent) in self
                    and self[str(file_path.parent)].is_overview
                )
                or (not file.is_documentation_page())
            ):
                if self._meta_plugin_config.slug.enabled:
                    # Get URL parts
                    if file.url.endswith("/"):
                        file.url = file.url[0:-1]
                    url_parts = file.url.split("/")

                    # Get abs file parts
                    path_parts: list[Path] = []
                    for path_part in file_path.parts:
                        if not path_parts:
                            path_parts.append(Path(path_part))
                        else:
                            path_parts.append(path_parts[-1] / path_part)

                    # Replace URL parts that have slug defined
                    for position, path_part in enumerate(path_parts):
                        meta_file: Optional[MetaFile] = self.get(str(path_part), None)
                        if meta_file is not None:
                            url_parts[position] = meta_file.slug

                    # Recreate file params based on URL with replaced parts
                    if file.url != ".":  # Do not modify main index page
                        file.url = quote(f"{'/'.join(url_parts)}/")
                        url_parts.append(file.dest_uri.split("/")[-1])
                        if len(url_parts) >= 2 and url_parts[-1] == url_parts[-2]:
                            url_parts.pop(-1)
                        file.dest_uri = quote("/".join(url_parts))
                        file.abs_dest_path = str(
                            Path(self._mkdocs_config.site_dir) / file.dest_uri
                        )
                new_files.append(file)
        return new_files

    def files_gen(self) -> Generator[MetaFile, Any, None]:
        for meta_file in self.values():
            meta_file: MetaFile
            yield meta_file



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from enum import Enum
from typing import Any
from typing import Optional
from typing import Union
from typing import cast

from mkdocs.config.base import Config
from mkdocs.config.defaults import MkDocsConfig

log = logging.getLogger("mkdocs.plugins.publisher._shared.mkdocs_utils")


class ConfigChoiceEnum(Enum):
    def __eq__(self, other) -> bool:
        if isinstance(other, bool) and self.is_bool:
            return self._str_to_bool(self.name) is other
        return self.name.lower() == str(other).lower()

    def __hash__(self):
        return super().__hash__()

    @staticmethod
    def _str_to_bool(text) -> bool:
        if text.lower() == "true":
            return True
        elif text.lower() == "false":
            return False
        raise ValueError(f"'{text}' cannot be converted into bool value")

    @classmethod
    def _get_enums(cls, enums: list) -> list:
        enums_list = []
        for enum in enums:
            if enum.is_bool:
                enums_list.extend([enum.name, cls._str_to_bool(enum.name)])
            else:
                enums_list.append(enum.name)
        return enums_list

    @property
    def name(self) -> str:
        return ".".join(str(self).split(".")[1:]).lower()

    @property
    def is_bool(self) -> bool:
        return self.value[2]

    @classmethod
    def default(cls) -> Optional[str]:
        defaults = [f.name for f in cls if f.value[1]]
        if len(defaults) == 0:
            return None
        elif len(defaults) > 1:
            raise ValueError(f"Multiple defaults specified: {defaults}")
        return defaults[0]

    @classmethod
    def choices(cls) -> list:
        return cls._get_enums(enums=cast(list, cls))


def get_plugin_config(
    mkdocs_config: MkDocsConfig, plugin_name: str
) -> Union[None, dict[str, Any], Config]:
    plugins = mkdocs_config.plugins
    if isinstance(plugins, list):
        for plugin in plugins:
            if isinstance(plugin, dict) and list(plugin.keys())[0] == plugin_name:
                return plugin[list(plugin.keys())[0]]
            elif isinstance(plugin, str) and plugin == plugin_name:
                return {}
        raise SystemError("Break")
    else:
        if plugin_name in mkdocs_config.plugins:
            return mkdocs_config.plugins[plugin_name].config
        else:
            return None


def get_mkdocs_config() -> MkDocsConfig:
    config = MkDocsConfig()
    with open("mkdocs.yml") as mkdocs_yml:
        config.load_file(mkdocs_yml)
    return cast(MkDocsConfig, config)



# MIT License
#
# Copyright (c) 2023-2024 Maciej 'maQ' Kusz <maciej.kusz@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from pathlib import Path
from typing import Optional

from mkdocs.config.defaults import MkDocsConfig
from mkdocs.plugins import BasePlugin
from mkdocs.plugins import event_priority
from mkdocs.structure.pages import Page

# noinspection PyProtectedMember
from mkdocs_publisher._shared.html_modifiers import HTMLModifier
from mkdocs_publisher.social.config import SocialConfig

log = logging.getLogger("mkdocs.plugins.publisher.social.plugin")

TWITTER_PROPERTIES = [
    "twitter:title",
    "twitter:description",
    "twitter:image",
    "twitter:card",
    "twitter:site",
    "twitter:creator",
]
OPEN_GRAPH_PROPERTIES = [
    "og:title",
    "og:description",
    "og:type",
    "og:url",
    "og:image",
    "og:site_name",
    "og:locale",
]


class SocialPlugin(BasePlugin[SocialConfig]):
    @event_priority(-99)
    def on_post_page(self, output: str, *, page: Page, config: MkDocsConfig) -> Optional[str]:
        html_modifier = HTMLModifier(markup=output)

        log.debug(f"Processing social properties for file: '{page.file.src_path}'")
        log.debug("Removing old Twitter and Open Graph properties")
        html_modifier.remove_meta_properties(properties=OPEN_GRAPH_PROPERTIES)
        html_modifier.remove_meta_properties(properties=TWITTER_PROPERTIES)

        # Get all needed meta values
        title = page.meta.get(self.config.meta_keys.title_key, None)
        description = page.meta.get(self.config.meta_keys.description_key, None)
        image = page.meta.get(self.config.meta_keys.image_key, None)
        if image is not None:
            image = str(image)
            # TODO: use obsidian path link solver when it will be developed
            if image.startswith("../"):
                image = f"/{image.replace('../', '')}"
            if image.startswith("/"):
                image = image[1:]
            image_path = Path(config.docs_dir) / Path(image)
            if not image_path.exists():
                log.warning(
                    f"File: '{str(image)}' doesn't exists!\n"
                    f"('{self.config.meta_keys.image_key}' meta key"
                    f" from '{page.file.src_path}' file.)"
                )
            image = f'{config.site_url}{image.replace("//", "/")}'
        url = f"{config.site_url}{page.url}"
        site_name = config.site_name

        if self.config.og.enabled and title and description:
            log.debug("Adding Open Graph properties")
            html_modifier.add_meta_property(name="og:type", value="article")
            html_modifier.add_meta_property(name="og:title", value=title)
            html_modifier.add_meta_property(name="og:description", value=description)
            html_modifier.add_meta_property(name="og:site_name", value=site_name)
            html_modifier.add_meta_property(name="og:locale", value=self.config.og.locale)
            html_modifier.add_meta_property(name="og:url", value=url)

            if image is not None:
                html_modifier.add_meta_property(name="og:image", value=image)

        if self.config.twitter.enabled and title and description:
            log.debug("Adding Twitter properties")
            card_type = "summary_large_image" if image else "summary"
            html_modifier.add_meta_property(name="twitter:card", value=card_type)
            html_modifier.add_meta_property(name="twitter:title", value=title)
            html_modifier.add_meta_property(name="twitter:description", value=description)

            if image is not None:
                html_modifier.add_meta_property(name="twitter:image", value=image)

            if self.config.twitter.website:
                html_modifier.add_meta_property(
                    name="twitter:site", value=self.config.twitter.website
                )

            if self.config.twitter.author:
                html_modifier.add_meta_property(
                    name="twitter:creator", value=self.config.twitter.author
                )

        return str(html_modifier)
